/***********************************************************
 *
 * This script (2 of 3) is part of the following presentation:
 *   Unique, Pseudo Random Numbers: Generate without using RAM, Disk, CPU, I/O, or even time
 *
 * Script:     UniquePseudoRandomNumbers_02_CreateObjects.sql
 * Created by: Solomon Rutzky  ( http://www.SqlQuantumLeap.com/ )
 * Company:    Sql Quantum Lift  ( http://www.SqlQuantumLift.com/ )
 * Created on: 2016-07-10
 * Updated on: 2017-06-19
 *
 * The latest version of this script can be found at:
 * http://pastebin.com/n6k8DeyC
 *
 ***********************************************************/
USE [tempdb];
SET NOCOUNT, ANSI_NULLS, ANSI_PADDING, QUOTED_IDENTIFIER ON;
---------------------------------------------------------------------------------
GO

--
----- This script IS intended to be run all at once (i.e. F5).
--
----- When connected to SQL Server, each section can be collapsed
----- (i.e. hidden) by clicking the "-" to the left of each "BEGIN".
--
GO

-----------------------------------------------------------------------
BEGIN -- Create Schema: [MMI]
IF (SCHEMA_ID(N'MMI') IS NULL)
BEGIN
	PRINT 'Creating [MMI] Schema...';
	EXEC (N'CREATE SCHEMA [MMI] AUTHORIZATION [dbo];');
END;

-----------------------------------------------------------------------
END
GO

-----------------------------------------------------------------------
BEGIN -- Create Function: [GreatestCommonDivisor]
IF (OBJECT_ID(N'MMI.GreatestCommonDivisor') IS NULL)
BEGIN
	PRINT 'Creating function stub for GreatestCommonDivisor...';
	EXEC (N'CREATE FUNCTION MMI.GreatestCommonDivisor () RETURNS INT AS BEGIN RETURN 0; END;');
END;

-----------------------------------------------------------------------
END
GO

PRINT 'Altering function GreatestCommonDivisor...';
GO

-- http://developer.classpath.org/doc/java/math/BigInteger-source.html
ALTER FUNCTION MMI.GreatestCommonDivisor(@NumeroUno BIGINT, @NumeroDos BIGINT)
RETURNS BIGINT
WITH SCHEMABINDING
AS
BEGIN
	DECLARE	@Temp BIGINT;

	IF (@NumeroUno < @NumeroDos)
	BEGIN
		SET @Temp = @NumeroUno;
		SET @NumeroUno = @NumeroDos;
		SET @NumeroDos = @Temp;
	END;

	WHILE (1 = 1)
	BEGIN
		IF (@NumeroDos = 0)
		BEGIN
			BREAK;
		END;

		IF (@NumeroDos = 1)
		BEGIN
			SET @NumeroUno = 1;
			BREAK;
			--RETURN 1;
		END;

		SET @Temp = @NumeroDos;
		SET @NumeroDos = @NumeroUno % @NumeroDos;
		SET @NumeroUno = @Temp;
	END;

	RETURN @NumeroUno;
END; 
-----------------------------------------------------------------------
GO

-----------------------------------------------------------------------
BEGIN -- Create Function: [ModularMultiplicativeInverse]
IF (OBJECT_ID(N'MMI.ModularMultiplicativeInverse') IS NULL)
BEGIN
	PRINT 'Creating function stub for ModularMultiplicativeInverse...';
	EXEC (N'CREATE FUNCTION MMI.ModularMultiplicativeInverse () RETURNS INT AS BEGIN RETURN 0; END;');
END;

-----------------------------------------------------------------------
END
GO

PRINT 'Altering function ModularMultiplicativeInverse...';
GO

-- http://rosettacode.org/wiki/Modular_inverse#C (second one)
ALTER FUNCTION MMI.ModularMultiplicativeInverse(@BaseValue BIGINT, @Modulo BIGINT)
RETURNS BIGINT
WITH SCHEMABINDING
AS
BEGIN
	DECLARE	@t BIGINT = 0,
			@nt BIGINT = 1,
			@q BIGINT,
			@tmp BIGINT;

	IF (@Modulo = 1)
	BEGIN
		RETURN 0;
	END;

	IF (@BaseValue = 1)
	BEGIN
		RETURN 1;
	END;

	SET @Modulo = ABS(@Modulo);
	IF (@BaseValue < 0)
	BEGIN
		SET @BaseValue = (@Modulo - (-@BaseValue % @Modulo));
	END;

	DECLARE	@r BIGINT = @Modulo,
			@nr BIGINT = (@BaseValue % @Modulo);

	WHILE (@nr <> 0)
	BEGIN
		SET @q = (@r / @nr);

		SET @tmp = @nt;
		SET @nt = (@t - (@q * @nt));
		SET @t = @tmp;

		SET @tmp = @nr;
		SET @nr = (@r % @nr);
		SET @r = @tmp;
	END;

	IF (@r > 1)
	BEGIN
		RETURN -1;
	END;

	IF (@t < 0)
	BEGIN
		SET @t = (@t + @Modulo);
	END;

	RETURN @t;
END;
-----------------------------------------------------------------------
GO

-----------------------------------------------------------------------
BEGIN -- Drop Function: [GenerateTest]
IF (OBJECT_ID(N'MMI.GenerateTest') IS NOT NULL)
BEGIN
	PRINT 'Dropping function GenerateTest...';
	EXEC (N'DROP FUNCTION MMI.GenerateTest;');
END;

-----------------------------------------------------------------------
END
GO

-----------------------------------------------------------------------
BEGIN -- Create Function: [Encode]
IF (OBJECT_ID(N'MMI.Encode') IS NULL)
BEGIN
	PRINT 'Creating function stub for Encode...';
	EXEC (N'CREATE FUNCTION MMI.Encode () RETURNS INT AS BEGIN RETURN 0; END;');
END;

-----------------------------------------------------------------------
END
GO

PRINT 'Altering function Encode...';
GO

-- https://ericlippert.com/2013/11/14/a-practical-use-of-multiplicative-inverses/
ALTER FUNCTION MMI.Encode(@BaseValue BIGINT, @Modulo BIGINT, @ValueToEncode BIGINT)
RETURNS BIGINT
WITH SCHEMABINDING
AS
BEGIN
	DECLARE @Result BIGINT = ((@ValueToEncode * @BaseValue) % @Modulo);

	RETURN CASE @Result WHEN 0 THEN @ValueToEncode ELSE @Result END;
END;
-----------------------------------------------------------------------
GO

-----------------------------------------------------------------------
BEGIN -- Create Function: [Decode]
IF (OBJECT_ID(N'MMI.Decode') IS NULL)
BEGIN
	PRINT 'Creating function stub for Decode...';
	EXEC (N'CREATE FUNCTION MMI.Decode () RETURNS INT AS BEGIN RETURN 0; END;');
END;

-----------------------------------------------------------------------
END
GO

PRINT 'Altering function Decode...';
GO

-- https://ericlippert.com/2013/11/14/a-practical-use-of-multiplicative-inverses/
ALTER FUNCTION MMI.Decode(@MultiplicativeInverse BIGINT, @Modulo BIGINT, @ValueToDecode BIGINT)
RETURNS BIGINT
WITH SCHEMABINDING
AS
BEGIN
	DECLARE @Result BIGINT = ((@ValueToDecode * @MultiplicativeInverse) % @Modulo);

	RETURN CASE @Result WHEN 0 THEN @ValueToDecode ELSE @Result END;
END;
-----------------------------------------------------------------------
GO

-----------------------------------------------------------------------
BEGIN -- Create Function: [GenerateTest]
IF (OBJECT_ID(N'MMI.GenerateTest') IS NULL)
BEGIN
	PRINT 'Creating function stub for GenerateTest...';
	EXEC (N'CREATE FUNCTION MMI.GenerateTest () RETURNS TABLE AS RETURN SELECT 1 AS [g];');
END;

-----------------------------------------------------------------------
END
GO

PRINT 'Altering function GenerateTest...';
GO

ALTER FUNCTION MMI.GenerateTest(@ValueCoprimeToUpperLimit BIGINT, @UpperLimit BIGINT,
							@MultiplicativeInverse BIGINT, @RowsToGenerate INT)
RETURNS TABLE
WITH SCHEMABINDING
AS RETURN
	WITH basenums(num) AS
	(
		SELECT	tmp.col
		FROM		(VALUES (0), (0), (0), (0), (0), (0), (0), (0), (0), (0),
				(0), (0), (0), (0), (0), (0), (0), (0), (0), (0),
				(0), (0), (0), (0), (0), (0), (0), (0), (0), (0),
				(0), (0), (0), (0), (0), (0), (0), (0), (0), (0)) tmp(col)
	), realnums(Num) AS
	(
		SELECT	TOP (@RowsToGenerate) ROW_NUMBER() OVER (ORDER BY (SELECT 0))
		FROM		basenums b1
		CROSS APPLY	basenums b2
		CROSS APPLY	basenums b3
	)
	SELECT	nm.[Num],
			MMI.Encode(@ValueCoprimeToUpperLimit, @UpperLimit, nm.[Num]) AS [Encoded],
			MMI.Decode(@MultiplicativeInverse, @UpperLimit,
				MMI.Encode(@ValueCoprimeToUpperLimit, @UpperLimit, nm.[Num])) AS [Decoded],
			MMI.Encode(@ValueCoprimeToUpperLimit, @UpperLimit, nm.[Num]) -
				(LAG(MMI.Encode(@ValueCoprimeToUpperLimit, @UpperLimit, nm.[Num])) OVER (ORDER BY nm.[Num]))
				AS [DifferenceFromPrior]
	FROM		realnums nm;
-----------------------------------------------------------------------
GO
